@@ -20,6 +20,8 @@
 *     [7] Quinlan, J. Ross. C4. 5: programs for machine learning. Elsevier, 2014.
 *     [8] Cortes, David. "Distance approximation using Isolation Forests." arXiv preprint arXiv:1910.12362 (2019).
 *     [9] Cortes, David. "Imputing missing values with unsupervised random trees." arXiv preprint arXiv:1911.06646 (2019).
+*     [10] https://math.stackexchange.com/questions/3333220/expected-average-depth-in-random-binary-tree-constructed-top-to-bottom
+*     [11] Cortes, David. "Revisiting randomized choices in isolation forests." arXiv preprint arXiv:2110.13402 (2021).
 * 
 *     BSD 2-Clause License
 *     Copyright (c) 2019-2021, David Cortes
@@ -48,6 +50,14 @@
    serialized raw bytes instead, as it will likely be faster due to better
    cache utilizations and those objects use less memory. */
 
+/* TODO: these trees are all created in a depth-first fashion, which will
+   not be cache-friendly when predictions are sent to a right-side branch. In
+   order to make predictions faster, could re-arrange the trees after-the-fact
+   so that they contain batches of consecutive nodes (parent and children and
+   grandchildren) up to some depth - that way these prediction functions would
+   run faster. After that, could also do a manual tree leaves unroll within each
+   batch with stack-assigned variables for an even faster prediction function. */
+
 
 /* Predict outlier score, average depth, or terminal node numbers
 * 
